# coding: utf-8
from test.util import build_grab
from test.util import BaseGrabTestCase


class GrabSimpleTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_get(self):
        self.server.response['get.data'] = 'Final Countdown'
        g = build_grab()
        g.go(self.server.get_url())
        self.assertTrue(b'Final Countdown' in g.response.body)

    def test_body_content(self):
        self.server.response['get.data'] = 'Simple String'
        g = build_grab()
        g.go(self.server.get_url())
        self.assertEqual(b'Simple String', g.response.body)
        # self.assertEqual('Simple String' in g.response.runtime_body)

    def test_status_code(self):
        self.server.response['get.data'] = 'Simple String'
        g = build_grab()
        g.go(self.server.get_url())
        self.assertEqual(200, g.response.code)

    def test_parsing_response_headers(self):
        self.server.response['headers'] = [('Hello', 'Grab')]
        g = build_grab()
        g.go(self.server.get_url())
        self.assertTrue(g.response.headers['Hello'] == 'Grab')

    def test_depreated_hammer_mode_options(self):
        self.server.response['get.data'] = 'foo'
        g = build_grab()
        g.setup(hammer_mode=True)
        g.go(self.server.get_url())

        g.setup(hammer_timeouts=((1, 1), (2, 2)))
        g.go(self.server.get_url())



# coding: utf-8
import os

from test.util import BaseGrabTestCase
from test.util import TMP_DIR, clear_directory, build_grab
from grab.base import reset_request_counter


class TestCookies(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_log_option(self):
        clear_directory(TMP_DIR)
        reset_request_counter()

        log_file_path = os.path.join(TMP_DIR, 'log.html')
        g = build_grab()
        g.setup(log_file=log_file_path)
        self.server.response['get.data'] = 'omsk'

        self.assertEqual(os.listdir(TMP_DIR), [])
        g.go(self.server.get_url())
        self.assertEqual(os.listdir(TMP_DIR), ['log.html'])
        self.assertEqual(open(log_file_path).read(), 'omsk')

    def test_log_dir_option(self):
        clear_directory(TMP_DIR)
        reset_request_counter()

        g = build_grab()
        g.setup(log_dir=TMP_DIR)
        self.server.response_once['get.data'] = 'omsk1'
        self.server.response['get.data'] = 'omsk2'

        self.assertEqual(os.listdir(TMP_DIR), [])
        g.go(self.server.get_url())
        g.go(self.server.get_url())
        self.assertEqual(sorted(os.listdir(TMP_DIR)),
                         ['01.html', '01.log', '02.html', '02.log'])
        self.assertEqual(open(os.path.join(TMP_DIR, '01.html')).read(),
                         'omsk1')
        self.assertEqual(open(os.path.join(TMP_DIR, '02.html')).read(),
                         'omsk2')

    def test_log_dir_response_content(self):
        clear_directory(TMP_DIR)
        reset_request_counter()

        g = build_grab()
        g.setup(log_dir=TMP_DIR)
        self.server.response['get.data'] = 'omsk'
        self.server.response['headers'] = [('X-Engine', 'PHP')]

        self.assertEqual(os.listdir(TMP_DIR), [])
        g.go(self.server.get_url())
        self.assertEqual(sorted(os.listdir(TMP_DIR)), ['01.html', '01.log'])
        log_file_content = open(os.path.join(TMP_DIR, '01.log')).read()
        self.assertTrue('X-Engine' in log_file_content)

    def test_log_dir_request_content_is_empty(self):
        clear_directory(TMP_DIR)
        reset_request_counter()

        g = build_grab()
        g.setup(log_dir=TMP_DIR)
        g.setup(headers={'X-Name': 'spider'}, post='xxxPost')

        self.assertEqual(os.listdir(TMP_DIR), [])
        g.go(self.server.get_url())
        self.assertEqual(sorted(os.listdir(TMP_DIR)), ['01.html', '01.log'])
        log_file_content = open(os.path.join(TMP_DIR, '01.log')).read()
        self.assertFalse('X-Name' in log_file_content)
        self.assertFalse('xxxPost' in log_file_content)

    def test_log_dir_request_content_headers_and_post(self):
        clear_directory(TMP_DIR)
        reset_request_counter()

        g = build_grab()
        g.setup(log_dir=TMP_DIR, debug=True)
        g.setup(headers={'X-Name': 'spider'}, post={'xxx': 'Post'})

        self.assertEqual(os.listdir(TMP_DIR), [])
        g.go(self.server.get_url())
        self.assertEqual(sorted(os.listdir(TMP_DIR)), ['01.html', '01.log'])
        log_file_content = open(os.path.join(TMP_DIR, '01.log')).read()
        self.assertTrue('X-Name' in log_file_content)
        self.assertTrue('xxx=Post' in log_file_content)

    def test_debug_post(self):
        g = build_grab(debug_post=True)
        g.setup(post={'foo': 'bar'})
        self.server.response['post.data'] = 'x'
        g.go(self.server.get_url())
        self.assertEqual(b'x', g.doc.body)

    def test_debug_post_integer_bug(self):
        g = build_grab(debug_post=True)
        g.setup(post={'foo': 3})
        self.server.response['post.data'] = 'x'
        g.go(self.server.get_url())
        self.assertEqual(b'x', g.doc.body)



from __future__ import absolute_import
import re
import itertools
import logging
from random import randint
import six
from collections import namedtuple

from grab.error import GrabError, GrabNetworkError

RE_SIMPLE_PROXY = re.compile(r'^([^:]+):([^:]+)$')
RE_AUTH_PROXY = re.compile(r'^([^:]+):([^:]+):([^:]+):([^:]+)$')
PROXY_FIELDS = ('host', 'port', 'username', 'password', 'proxy_type')
logger = logging.getLogger('grab.proxylist')


class Proxy(namedtuple('Proxy', PROXY_FIELDS)):
    def get_address(self):
        return '%s:%s' % (self.host, self.port)

    def get_userpwd(self):
        if self.username:
            return '%s:%s' % (self.username, self.password or '')


class InvalidProxyLine(GrabError):
    pass


def parse_proxy_line(line):
    """
    Parse proxy details from the raw text line.

    The text line could be in one of the following formats:
    * host:port
    * host:port:username:password
    """

    line = line.strip()
    match = RE_SIMPLE_PROXY.search(line)
    if match:
        return match.group(1), match.group(2), None, None

    match = RE_AUTH_PROXY.search(line)
    if match:
        host, port, user, pwd = match.groups()
        return host, port, user, pwd

    raise InvalidProxyLine('Invalid proxy line: %s' % line)


def parse_raw_list_data(data, proxy_type='http'):
    "Iterate over proxy servers found in the raw data"
    if not isinstance(data, six.text_type):
        data = data.decode('utf-8')
    for orig_line in data.splitlines():
        line = orig_line.strip().replace(' ', '')
        if line and not line.startswith('#'):
            try:
                host, port, username, password = parse_proxy_line(line)
            except InvalidProxyLine as ex:
                logger.error(ex)
            else:
                yield Proxy(host, port, username, password, proxy_type)


class BaseProxySource(object):
    def __init__(self, **kwargs):
        kwargs.setdefault('proxy_type', 'http')
        self.config = kwargs

    def load(self):
        data = self.load_raw_data()
        return list(parse_raw_list_data(
            data, proxy_type=self.config['proxy_type']))


class FileProxySource(BaseProxySource):
    "Proxy source that loads list from the file"
    def __init__(self, path, **kwargs):
        self.path = path
        super(FileProxySource, self).__init__(**kwargs)

    def load_raw_data(self):
        return open(self.path).read()


class WebProxySource(BaseProxySource):
    "Proxy source that loads list from web resource"
    def __init__(self, url, **kwargs):
        self.url = url
        super(WebProxySource, self).__init__(**kwargs)

    def load_raw_data(self):
        from grab import Grab
        limit = 3
        for count in range(limit):
            try:
                return Grab().go(url=self.url).unicode_body()
            except GrabNetworkError:
                if count >= (limit - 1):
                    raise


class ProxyList(object):
    """
    Class to work with proxy list.
    """

    def __init__(self, source=None):
        self._source = source
        self._list = []
        self._list_iter = None

    def set_source(self, source):
        "Set the proxy source and use it to load proxy list"
        self._source = source
        self.load()

    def load_file(self, path, proxy_type='http'):
        "Load proxy list from file"
        self.set_source(FileProxySource(path, proxy_type=proxy_type))

    def load_url(self, url, proxy_type='http'):
        "Load proxy list from web document"
        self.set_source(WebProxySource(url, proxy_type=proxy_type))

    def load(self):
        "Load proxy list from configured proxy source"
        self._list = self._source.load()
        self._list_iter = itertools.cycle(self._list)

    def get_random_proxy(self):
        "Return random proxy"
        idx = randint(0, len(self._list) - 1)
        return self._list[idx]

    def get_next_proxy(self):
        "Return next proxy"
        return self._list_iter.next()

    def size(self):
        "Return number of proxies in the list"
        return len(self._list)



from unittest import TestCase

from test.util import get_temp_file
from grab.export.mysql_dumper import MysqlCSVDumper

class MysqlCSVDumperTestCase(TestCase):
    def test_constructor(self):
        path = get_temp_file()
        dumper = MysqlCSVDumper(path)



# coding: utf-8
from __future__ import absolute_import
import os.path

from test.util import TEST_DIR, TMP_DIR, build_grab
from test.util import BaseGrabTestCase

HTML = """
Hello world
"""

IMG_FILE = os.path.join(TEST_DIR, 'files', 'yandex.png')


class TestResponse(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_save(self):
        "Test `Response.save` method."
        img_data = open(IMG_FILE, 'rb').read()
        temp_file = os.path.join(TMP_DIR, 'file.bin')
        self.server.response['get.data'] = img_data

        g = build_grab()
        g.go(self.server.get_url())
        g.response.save(temp_file)
        self.assertEqual(open(temp_file, 'rb').read(), img_data)

    def test_save_hash(self):
        "Test `Response.save_hash` method."
        img_data = open(IMG_FILE, 'rb').read()
        self.server.response['get.data'] = img_data

        g = build_grab()
        g.go(self.server.get_url())
        path = g.response.save_hash(self.server.get_url(), TMP_DIR)
        test_data = open(os.path.join(TMP_DIR, path), 'rb').read()
        self.assertEqual(test_data, img_data)

    def test_custom_charset(self):
        self.server.response['get.data'] = u'<html><head><meta '\
            u'http-equiv="Content-Type" content="text/html; '\
            u'charset=utf8;charset=cp1251" /></head><body>'\
            u'<h1>крокодил</h1></body></html>'.encode('utf-8')
        g = build_grab()
        g.setup(document_charset='utf-8')
        g.go(self.server.get_url())
        self.assertTrue(u'крокодил' in g.response.unicode_body())

    def test_xml_declaration(self):
        """
        unicode_body() should return HTML with xml declaration (if it
        exists in original HTML)
        """
        self.server.response['get.data'] = """<?xml version="1.0" encoding="UTF-8"?>
        <html><body><h1>тест</h1></body></html>
        """
        g = build_grab()
        g.go(self.server.get_url())
        ubody = g.response.unicode_body()
        self.assertTrue(u'тест' in ubody)
        self.assertTrue('<?xml' in ubody)



from unittest import TestCase
from importlib import import_module
from grab.tools.hook import CustomImporter

MODULES = """
content
feed
http
logs
ping
russian
text
work
control
files
__init__
lxml_tools
progress
selenium_tools
user_agent
yandex
debug
google
internal
metric
pwork
structured
w3lib_encoding
encoding
html
lock
parser
rex
system
watch
""".strip().splitlines()

class ToolsDeprecatedTestCase(TestCase):
    def setUp(self):
        self.hook = CustomImporter()

    def test_import_traditional_way(self):
        for mod_name in MODULES:
            path = 'grab.tools.%s' % mod_name
            import_module(path)

    def test_hook(self):
        for mod_name in MODULES:
            path = 'grab.tools.%s' % mod_name
            self.hook.find_module(path)
            self.hook.load_module(path)

    def test_hook2(self):
        self.assertIsInstance(self.hook.find_module('grab.tools.yandex'),
                              CustomImporter)
        self.assertEqual(self.hook.find_module('grab.spider'),
                         None)

        self.hook.find_module('grab.tools.lxml_tools')
        self.assertEqual(self.hook.name, '.etree')

        self.hook.find_module('grab.tools.module')
        self.assertRaises(ImportError, self.hook.load_module,
                          'grab.tools.module')




# coding: utf-8
from json import loads

from weblib.structured import Structure as x

from test.util import build_grab
from test.util import BaseGrabTestCase


XML = b'''
    <issue index="2">
        <title>XML today</title>
        <date>12.09.98</date>
        <about>XML</about>
        <home-url>www.j.ru/issues/</home-url>
        <number>448</number>
        <detail>
            <description>

                issue 2
                detail description

            </description>
            <number>445</number>
        </detail>
        <articles>
            <article ID="3">
                <title>Issue overview</title>
                <url>/article1</url>
                <hotkeys>
                    <hotkey>language</hotkey>
                    <hotkey>marckup</hotkey>
                    <hotkey>hypertext</hotkey>
                </hotkeys>
                <article-finished/>
            </article>
            <article>
                <title>Latest reviews</title>
                <url>/article2</url>
                <author ID="3"/>
                <hotkeys>
                    <hotkey/>
                </hotkeys>
            </article>
            <article ID="4">
                <title/>
                <url/>
                <hotkeys/>
            </article>
        </articles>
    </issue>
'''


class StructuredExtensionTest(BaseGrabTestCase):
    def setUp(self):
        self.g = build_grab(document_body=XML)

    def test_1(self):
        result = self.g.doc.structure(
            '//issue',
            title='./title/text()',
            date='./date/text()',
            about='./about/text()',
            number=('./number/text()', int),
            home_url='./home-url/text()',
        )
        self.assertEqual(
            result,
            loads('''
                [
                    {
                        "date": "12.09.98",
                        "about": "XML",
                        "home_url": "www.j.ru/issues/",
                        "number": 448,
                        "title": "XML today"
                    }
                ]
            ''')
        )

    def test_2(self):
        result = self.g.doc.structure(
            '//issue',
            x(
                './detail',
                description=('./description/text()',
                             lambda item: ' '.join(item.split())),
                detail_number=('./number/text()', int)
            ),
            title='./title/text()',
            date='./date/text()',
        )
        self.assertEqual(
            result,
            loads('''
                [
                    {
                        "detail_number": 445,
                        "date": "12.09.98",
                        "description": "issue 2 detail description",
                        "title": "XML today"
                    }
                ]
            ''')
        )

    def test_3(self):
        result = self.g.doc.structure(
            '//issue',
            home_url='./home-url/text()',
            articles=x(
                './articles/article',
                id='./@id',
                title='./title/text()',
                url='./url/text()',
                hotkeys=x(
                    './hotkeys',
                    hotkey='./hotkey/text()'
                )
            )
        )
        self.assertEqual(
            result,
            loads('''
                [
                    {
                        "articles": [
                            {
                                "url": "/article1",
                                "id": "3",
                                "hotkeys": [
                                    {
                                        "hotkey": "language"
                                    }
                                ],
                                "title": "Issue overview"
                            },
                            {
                                "url": "/article2",
                                "id": null,
                                "hotkeys": [
                                    {
                                        "hotkey": null
                                    }
                                ],
                                "title": "Latest reviews"
                            },
                            {
                                "url": null,
                                "id": "4",
                                "hotkeys": [
                                    {
                                        "hotkey": null
                                    }
                                ],
                                "title": null
                            }
                        ],
                        "home_url": "www.j.ru/issues/"
                    }
                ]
            ''')
        )



import logging
import six

from grab.spider.base import Spider
from grab.spider.error import SpiderInternalError
from grab.util.config import build_root_config, build_spider_config

SPIDER_REGISTRY = {}
logger = logging.getLogger('grab.util.module')


def build_spider_registry(config):
    SPIDER_REGISTRY.clear()

    opt_modules = []
    opt_modules = config['global'].get('spider_modules', [])

    for path in opt_modules:
        if ':' in path:
            path, cls_name = path.split(':')
        else:
            cls_name = None
        try:
            mod = __import__(path, None, None, ['foo'])
        except ImportError as ex:
            if path not in six.text_type(ex):
                logging.error('', exc_info=ex)
        else:
            for key in dir(mod):
                if key == 'Spider':
                    continue
                if cls_name is None or key == cls_name:
                    val = getattr(mod, key)
                    if isinstance(val, type) and issubclass(val, Spider):
                        if val.Meta.abstract:
                            pass
                        else:
                            spider_name = val.get_spider_name()
                            logger.debug(
                                'Module `%s`, found spider `%s` '
                                'with name `%s`' % (
                                    path, val.__name__, spider_name))
                            if spider_name in SPIDER_REGISTRY:
                                mod = SPIDER_REGISTRY[spider_name].__module__
                                raise SpiderInternalError(
                                    'There are two different spiders with '
                                    'the same name "%s". '
                                    'Modules: %s and %s' % (
                                        spider_name, mod, val.__module__))
                            else:
                                SPIDER_REGISTRY[spider_name] = val
    return SPIDER_REGISTRY


def load_spider_class(config, spider_name):
    if not SPIDER_REGISTRY:
        build_spider_registry(config)
    if spider_name not in SPIDER_REGISTRY:
        raise SpiderInternalError('Unknown spider: %s' % spider_name)
    else:
        return SPIDER_REGISTRY[spider_name]


def build_spider_instance(cls, settings_module, **kwargs):
    root_config = build_root_config(settings_module)
    spider_config = build_spider_config(cls, root_config)
    return cls(config=spider_config)



from __future__ import absolute_import
import logging
from weblib.etree import get_node_text
from weblib.text import find_number
from weblib.const import NULL
from weblib.error import DataNotFound
from weblib.encoding import make_unicode
import six

from grab.util.misc import deprecated
from grab.error import GrabMisuseError
from grab import error


class DeprecatedThings(object):
    __slots__ = ()

    """
    This super-class contains all deprecated things that are
    still in Grab class for back-ward compatibility.
    """

    # Deprecated methods from grab.ext.text module
    # ********************************************

    @deprecated(use_instead='grab.doc.text_search')
    def search(self, anchor, byte=False):
        return self.doc.text_search(anchor, byte=byte)

    @deprecated(use_instead='grab.doc.text_assert')
    def assert_substring(self, anchor, byte=False):
        return self.doc.text_assert(anchor, byte=byte)

    @deprecated(use_instead='grab.doc.text_assert_any')
    def assert_substrings(self, anchors, byte=False):
        return self.doc.text_assert_any(anchors, byte=byte)

    # Deprecated methods from grab.ext.rex module
    # ********************************************

    @deprecated(use_instead='grab.doc.rex_text')
    def rex_text(self, regexp, flags=0, byte=False, default=NULL):
        return self.doc.rex_text(regexp, flags=flags,
                                 byte=byte, default=default)

    @deprecated(use_instead='grab.doc.rex_search')
    def rex(self, regexp, flags=0, byte=False, default=NULL):
        return self.doc.rex_search(regexp, flags=flags,
                                   byte=byte, default=default)

    @deprecated(use_instead='grab.doc.rex_assert')
    def assert_rex(self, regexp, byte=False):
        return self.doc.rex_assert(regexp, byte=byte)

    # Deprecated methods from grab.ext.lxml
    # *************************************

    @property
    @deprecated(use_instead='grab.doc.tree')
    def tree(self):
        return self.doc.tree

    @deprecated(use_instead='grab.doc.build_html_tree')
    def build_html_tree(self):
        return self.doc.build_html_tree()

    @property
    @deprecated(use_instead='grab.doc.xml_tree')
    def xml_tree(self):
        return self.doc.xml_tree

    @deprecated(use_instead='grab.doc.build_xml_tree()')
    def build_xml_tree(self):
        return self.doc.build_xml_tree()

    @deprecated()
    def find_link(self, href_pattern, make_absolute=True):
        """
        Find link in response body which href value matches ``href_pattern``.

        Returns found url or None.
        """

        if make_absolute:
            self.tree.make_links_absolute(self.response.url)

        if isinstance(href_pattern, six.text_type):
            raise GrabMisuseError('Method `find_link` accepts only '
                                  'byte-string argument')
        href_pattern = make_unicode(href_pattern)
        for elem, attr, link, pos in self.tree.iterlinks():
            if elem.tag == 'a' and href_pattern in link:
                return link
        return None

    @deprecated()
    def find_link_rex(self, rex, make_absolute=True):
        """
        Find link matched the given regular expression in response body.

        Returns found url or None.
        """

        if make_absolute:
            self.tree.make_links_absolute(self.response.url)

        for elem, attr, link, pos in self.tree.iterlinks():
            if elem.tag == 'a':
                match = rex.search(link)
                if match:
                    # That does not work for string object
                    # link.match = match
                    return link
        return None

    @deprecated(use_instead='grab.doc.select().node()')
    def xpath(self, path, default=NULL, filter=None):
        if filter is not None:
            raise GrabMisuseError('Argument `filter` is not supported anymore')
        return self.doc.select(path).node(default=default)

    @deprecated(use_instead='grab.doc.select().one()')
    def xpath_one(self, path, default=NULL, filter=None):
        if filter is not None:
            raise GrabMisuseError('Argument `filter` is not supported anymore')
        return self.doc.select(path).node(default=default)

    @deprecated(use_instead='grab.doc.select()')
    def xpath_list(self, path, filter=None):
        if filter is not None:
            raise GrabMisuseError('Argument `filter` is not supported anymore')
        return self.doc.select(path).node_list()

    @deprecated(use_instead='grab.doc.select().text()')
    def xpath_text(self, path, default=NULL, filter=None, smart=False,
                   normalize_space=True):
        if filter is not None:
            raise GrabMisuseError('Argument `filter` is not supported anymore')
        return self.doc.select(path).text(default=default, smart=smart,
                                          normalize_space=normalize_space)

    @deprecated(use_instead='grab.doc.select().number()')
    def xpath_number(self, path, default=NULL, filter=None,
                     ignore_spaces=False,
                     smart=False, make_int=True):

        if filter is not None:
            raise GrabMisuseError('Argument `filter` is not supported anymore')
        return self.doc.select(path).number(default=default, smart=smart,
                                            ignore_spaces=ignore_spaces,
                                            make_int=make_int)

    @deprecated(use_instead='grab.doc.select().exists()')
    def xpath_exists(self, path):
        return self.doc.select(path).exists()

    @deprecated()
    def css(self, *args, **kwargs):
        return self.css_one(*args, **kwargs)

    @deprecated()
    def css_one(self, path, default=NULL):
        """
        Get first element which matches the given css path
            or raise DataNotFound.
        """

        try:
            return self.css_list(path)[0]
        except IndexError:
            if default is NULL:
                raise DataNotFound('CSS path not found: %s' % path)
            else:
                return default

    @deprecated()
    def css_list(self, path):
        """
        Find all elements which match given css path.
        """

        return self.tree.cssselect(path)

    @deprecated()
    def css_text(self, path, default=NULL, smart=False, normalize_space=True):
        """
        Get normalized text of node which matches the css path.
        """

        try:
            return get_node_text(self.css_one(path), smart=smart,
                                 normalize_space=normalize_space)
        except IndexError:
            if default is NULL:
                raise
            else:
                return default

    @deprecated()
    def css_number(self, path, default=NULL, ignore_spaces=False, smart=False,
                   make_int=True):
        """
        Find number in normalized text of node which
            matches the given css path.
        """

        try:
            text = self.css_text(path, smart=smart)
            return find_number(text, ignore_spaces=ignore_spaces,
                               make_int=make_int)
        except IndexError:
            if default is NULL:
                raise
            else:
                return default

    @deprecated()
    def assert_css(self, path):
        """
        If css path is not found then raise `DataNotFound` exception.
        """

        self.css_one(path)

    @deprecated()
    def assert_xpath(self, path):
        """
        If xpath path is not found then raise `DataNotFound` exception.
        """

        self.xpath_one(path)

    @deprecated()
    def css_exists(self, path):
        """
        Return True if at least one element with specified css path exists.
        """

        return len(self.css_list(path)) > 0

    @deprecated()
    def strip_tags(self, content, smart=False):
        """
        Strip tags from the HTML content.
        """
        from lxml.html import fromstring

        return get_node_text(fromstring(content), smart=smart)

    # Methods from deprecated grab.ext.pquery module
    # **********************************************

    @deprecated(use_instead='grab.doc.pyquery()')
    def pyquery(self, query):
        return self.doc.pyquery(query)

    # Response related things
    # ***********************

    # Backward compat.
    def _get_response(self):
        return self.doc

    def _set_response(self, val):
        self.doc = val

    response = property(_get_response, _set_response)

    @deprecated(use_instead='grab.setup_document')
    def fake_response(self, *args, **kwargs):
        return self.setup_document(*args, **kwargs)

    # Cookies
    # *******
    @deprecated(use_instead='grab.cookies.load_from_file')
    def load_cookies(self, path, file_required=True):
        self.cookies.load_from_file(path)

    @deprecated(use_instead='grab.cookies.save_to_file')
    def dump_cookies(self, path):
        self.cookies.save_to_file(path)

    @deprecated(use_instead='grab.proxylist.load_file OR '
                            'grab.proxylist.load_url')
    def load_proxylist(self, source, source_type, proxy_type='http',
                       auto_init=True, auto_change=True,
                       **kwargs):
        # self.proxylist = ProxyList(source, source_type,
        #                            proxy_type=proxy_type, **kwargs)
        if source_type == 'text_file':
            self.proxylist.load_file(source, proxy_type=proxy_type)
        elif source_type == 'url':
            self.proxylist.load_url(source, proxy_type=proxy_type)
        else:
            raise error.GrabMisuseError(
                'Unknown proxy source type: %s' % source_type)

        # self.proxylist.setup(auto_change=auto_change, auto_init=auto_init)
        self.setup(proxy_auto_change=auto_change)
        if not auto_change and auto_init:
            self.change_proxy()

    # Methods from deprecated grab.ext.form module
    # **********************************************

    @deprecated(use_instead='grab.doc.choose_form')
    def choose_form(self, number=None, id=None, name=None, xpath=None):
        return self.doc.choose_form(number=number, id=id,
                                    name=name, xpath=xpath)

    @property
    def form(self):
        logging.error('This attribut is deprecated. '
                      'Use grab.doc.form instead.')
        return self.doc.form

    @deprecated(use_instead='grab.doc.set_input')
    def set_input(self, name, value):
        return self.doc.set_input(name, value)

    @deprecated(use_instead='grab.doc.set_input_by_id')
    def set_input_by_id(self, _id, value):
        return self.doc.set_input_by_id(_id, value)

    @deprecated(use_instead='grab.doc.set_input_by_number')
    def set_input_by_number(self, number, value):
        return self.doc.set_input_by_number(number, value)

    @deprecated(use_instead='grab.doc.set_input_by_xpath')
    def set_input_by_xpath(self, xpath, value):
        return self.doc.set_input_by_xpath(xpath, value)

    @deprecated(use_instead='grab.doc.submit')
    def submit(self, submit_name=None, make_request=True,
               url=None, extra_post=None):
        return self.doc.submit(submit_name=submit_name,
                               make_request=make_request,
                               url=url, extra_post=extra_post)

    @deprecated(use_instead='grab.doc.form_fields')
    def form_fields(self):
        return self.doc.form_fields()

    @deprecated(use_instead='grab.doc.choose_form_by_element')
    def choose_form_by_element(self, xpath):
        return self.doc.choose_form_by_element(xpath)



import six
from grab.spider import Spider, Task
from grab.spider.error import SpiderError, FatalError
import os
import signal
import mock

from test.util import BaseGrabTestCase, build_spider


class BasicSpiderTestCase(BaseGrabTestCase):
    class SimpleSpider(Spider):
        def task_baz(self, grab, task):
            self.stat.collect('SAVED_ITEM', grab.response.body)

    def setUp(self):
        self.server.reset()

    def test_spider(self):
        self.server.response['get.data'] = 'Hello spider!'
        self.server.response['sleep'] = 0
        bot = build_spider(self.SimpleSpider)
        bot.setup_queue()
        bot.add_task(Task('baz', self.server.get_url()))
        bot.run()
        self.assertEqual(b'Hello spider!',
                         bot.stat.collections['SAVED_ITEM'][0])

    def test_network_limit(self):
        self.server.response['get.data'] = 'Hello spider!'
        self.server.response['sleep'] = 1.1

        bot = build_spider(self.SimpleSpider, network_try_limit=1)
        bot.setup_queue()
        bot.setup_grab(connect_timeout=1, timeout=1)
        bot.add_task(Task('baz', self.server.get_url()))
        bot.run()
        self.assertEqual(bot.stat.counters['spider:request-network'], 1)

        bot = build_spider(self.SimpleSpider, network_try_limit=2)
        bot.setup_queue()
        bot.setup_grab(connect_timeout=1, timeout=1)
        bot.add_task(Task('baz', self.server.get_url()))
        bot.run()
        self.assertEqual(bot.stat.counters['spider:request-network'], 2)

    def test_task_limit(self):
        self.server.response['get.data'] = 'Hello spider!'
        self.server.response['sleep'] = 1.1

        bot = build_spider(self.SimpleSpider, network_try_limit=1)
        bot.setup_grab(connect_timeout=1, timeout=1)
        bot.setup_queue()
        bot.add_task(Task('baz', self.server.get_url()))
        bot.run()
        self.assertEqual(bot.stat.counters['spider:task-baz'], 1)

        bot = build_spider(self.SimpleSpider, task_try_limit=2)
        bot.setup_queue()
        bot.add_task(Task('baz', self.server.get_url(), task_try_count=3))
        bot.run()
        self.assertEqual(bot.stat.counters['spider:request-network'], 0)

    def test_task_retry(self):
        self.server.response['get.data'] = 'xxx'
        self.server.response_once['code'] = 403
        bot = build_spider(self.SimpleSpider)
        bot.setup_queue()
        bot.add_task(Task('baz', self.server.get_url()))
        bot.run()
        self.assertEqual(b'xxx', bot.stat.collections['SAVED_ITEM'][0])

    def test_setup_grab(self):
        # Mulitple calls to `setup_grab` should accumulate
        # changes in config object.
        bot = build_spider(self.SimpleSpider)
        bot.setup_grab(log_dir='/tmp')
        bot.setup_grab(timeout=30)
        grab = bot.create_grab_instance()
        self.assertEqual(grab.config['log_dir'], '/tmp')
        self.assertEqual(grab.config['timeout'], 30)

    def test_generator(self):
        server = self.server

        class TestSpider(Spider):
            def task_generator(self):
                for x in six.moves.range(1111):
                    yield Task('page', url=server.get_url())

            def task_page(self, grab, task):
                self.stat.inc('count')

        bot = build_spider(TestSpider)
        bot.run()
        self.assertEqual(bot.stat.counters['count'], 1111)

    def test_get_spider_name(self):
        class TestSpider(Spider):
            pass

        self.assertEqual('test_spider', TestSpider.get_spider_name())

        class TestSpider(Spider):
            spider_name = 'foo_bar'

        self.assertEqual('foo_bar', TestSpider.get_spider_name())

    def test_handler_result_none(self):
        class TestSpider(Spider):
            def prepare(self):
                self.points = []

            def task_page(self, grab, task):
                yield None

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()

    """
    # DOES NOT WORK
    def test_keyboard_interrupt(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                os.kill(os.getpid(), signal.SIGINT)

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertTrue(bot.interrupted)
    """

    def test_fallback_handler_by_default_name(self):
        class TestSpider(Spider):
            def prepare(self):
                self.points = []

            def task_page(self, grab, task):
                pass

            def task_page_fallback(self, task):
                self.points.append(1)

        self.server.response['code'] = 403

        bot = build_spider(TestSpider, network_try_limit=1)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertEquals(bot.points, [1])

    def test_fallback_handler_by_fallback_name(self):
        class TestSpider(Spider):
            def prepare(self):
                self.points = []

            def task_page(self, grab, task):
                pass

            def fallback_zz(self, task):
                self.points.append(1)

        self.server.response['code'] = 403

        bot = build_spider(TestSpider, network_try_limit=1)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url(),
                          fallback_name='fallback_zz'))
        bot.run()
        self.assertEquals(bot.points, [1])

    #AFTER THIS: BAD
    def test_check_task_limits_invalid_value(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                pass

            def check_task_limits(self, task):
                return False, 'zz'

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url(),
                          fallback_name='fallback_zz'))
        self.assertRaises(SpiderError, bot.run)

    def test_handler_result_invalid(self):
        class TestSpider(Spider):
            def prepare(self):
                self.points = []

            def task_page(self, grab, task):
                yield 1

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        #bot.run()
        #self.assertEqual(1, bot.stat.counters['spider:error-spidererror'])
        self.assertRaises(SpiderError, bot.run)

    def test_fatal_error(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                raise FatalError

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        self.assertRaises(FatalError, bot.run)



from copy import deepcopy
from grab.util.default_config import DEFAULT_SPIDER_GLOBAL_CONFIG
import importlib


def update_dict(target, source, overwrite=False):
    for key, val in source.items():
        if key not in target or overwrite:
            target[key] = deepcopy(source[key])
    return target


def build_root_config(settings_mod_path):
    module = importlib.import_module(settings_mod_path)
    config = module.GRAB_SPIDER_CONFIG
    if not 'global' in config:
        config['global'] = {}
    update_dict(config['global'], DEFAULT_SPIDER_GLOBAL_CONFIG,
                overwrite=False)
    return config


def build_spider_config(spider_class, root_config):
    spider_name = spider_class.get_spider_name()
    spider_config = deepcopy(root_config.get(spider_name, {}))
    update_dict(spider_config, root_config['global'], overwrite=False)
    spider_class.setup_spider_config(spider_config)
    return spider_config



from datetime import datetime
try:
    from Queue import PriorityQueue, Empty
except ImportError:
    from queue import PriorityQueue, Empty

from grab.spider.queue_backend.base import QueueInterface


class QueueBackend(QueueInterface):
    def __init__(self, spider_name, **kwargs):
        super(QueueInterface, self).__init__(**kwargs)
        self.queue_object = PriorityQueue()
        self.schedule_list = []

    def put(self, task, priority, schedule_time=None):
        if schedule_time is None:
            self.queue_object.put((priority, task))
        else:
            self.schedule_list.append((schedule_time, task))

    def get(self):
        now = datetime.utcnow()

        removed_indexes = []
        index = 0
        for schedule_time, task in self.schedule_list:
            if schedule_time <= now:
                self.put(task, 1)
                removed_indexes.append(index)
            index += 1

        self.schedule_list = [x for idx, x in enumerate(self.schedule_list)
                              if idx not in removed_indexes]

        priority, task = self.queue_object.get(block=False)
        return task

    def size(self):
        return self.queue_object.qsize() + len(self.schedule_list)

    def clear(self):
        try:
            while True:
                self.queue_object.get(False)
        except Empty:
            pass
        self.schedule_list = []



from unittest import TestCase
from importlib import import_module
from lxml.html import fromstring
from six import StringIO
import mock

from grab.selector import XpathSelector

MODULES = (
    'grab.selector',
    'grab.selector.selector',
)

class ToolsDeprecatedTestCase(TestCase):
    def test(self):
        for path in MODULES:
            import_module(path)

    def test_xpath_selector(self):
        tree = fromstring('<div>test</div>')
        out = StringIO()
        with mock.patch('sys.stderr', out):
            XpathSelector(tree)
        self.assertTrue(
            'using XpathSelector from deprecated' in out.getvalue())



"""
QueueInterface defines interface of queue backend.
"""


class QueueInterface(object):
    def __init__(self, spider_name, **kwargs):
        pass

    def put(self, task, priority):
        raise NotImplementedError

    def get(self):
        """
        Return `Task` object or raise `Queue.Empty` exception

        @returns: `grab.spider.task.Task` object
        @raises: `Queue.Empty` exception
        """
        raise NotImplementedError

    def size(self):
        raise NotImplementedError

    def clear(self):
        """Remove all tasks from the queue."""
        raise NotImplementedError



import os
import logging
import shutil
import re
from grab.error import GrabError

logger = logging.getLogger('grab.script.start_project')


def setup_arg_parser(parser):
    parser.add_argument('project_name')
    parser.add_argument('--template')


def process_macros(content, context):
    changed = False
    for key, value in context.items():
        re_macros = re.compile(r'\{\{\s*%s\s*\}\}' % re.escape(key))
        if re_macros.search(content):
            changed = True
            content = re_macros.sub(value, content)
    return changed, content


def underscore_to_camelcase(val):
    items = val.lower().split('_')
    return ''.join(x.title() for x in items)


def normalize_extension(fname):
    if fname.endswith('._py'):
        return fname[:-4] + '.py'
    else:
        return fname


def main(project_name, template, **kwargs):
    cur_dir = os.getcwd()
    project_dir = os.path.join(cur_dir, project_name)

    if template is None:
        grab_root = os.path.dirname(
            os.path.dirname(os.path.realpath(__file__)))
        template_path = os.path.join(grab_root, 'util/default_project')
    else:
        template_path = template

    if os.path.exists(project_dir):
        raise GrabError('Directory %s already exists' % project_dir)
    else:
        logger.debug('Copying %s to %s' % (template_path, project_dir))
        shutil.copytree(template_path, project_dir)

        project_name_camelcase = underscore_to_camelcase(project_name)
        context = {
            'PROJECT_NAME': project_name,
            'PROJECT_NAME_CAMELCASE': project_name_camelcase,
        }
        for base, dir_names, file_names in os.walk(project_dir):
            for file_name in file_names:
                if file_name.endswith(('._py', '.py')):
                    file_path = os.path.join(base, file_name)
                    changed, content = process_macros(open(file_path).read(),
                                                      context)
                    if changed:
                        with open(normalize_extension(file_path), 'w') as out:
                            out.write(content)
                        os.unlink(file_path)



from grab.spider import Spider, Task, Data, NoDataHandler

from test.util import BaseGrabTestCase, build_spider


class TestSpider(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_data_nohandler_error(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                yield Data('foo', num=1)

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        self.assertRaises(NoDataHandler, bot.run)

    def test_exception_from_data_handler(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                yield Data('foo', num=1)

            def data_foo(self, num):
                1/0

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertTrue('data_foo' in bot.stat.collections['fatal'][0])

    def test_data_simple_case(self):
        class TestSpider(Spider):
            def prepare(self):
                self.data_processed = []

            def task_page(self, grab, task):
                yield Data('foo', number=1)

            def data_foo(self, number):
                self.data_processed.append(number)

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertEqual(bot.data_processed, [1])

    def test_complex_data(self):
        class TestSpider(Spider):
            def prepare(self):
                self.data_processed = []

            def task_page(self, grab, task):
                yield Data('foo', one=1, two=2, bar='gaz')

            def data_foo(self, one, two, **kwargs):
                self.data_processed.append(one)
                self.data_processed.append(two)
                self.data_processed.append(kwargs)

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertEqual(bot.data_processed, [1, 2, {'bar': 'gaz'}])

    def test_data_object_dict_interface(self):
        data = Data('person', person={'age': 22})
        self.assertRaises(KeyError, lambda: data['name'])
        self.assertEqual(data['person'], {'age': 22})

    def test_data_object_get_method(self):
        data = Data('person', person={'age': 22})
        self.assertRaises(KeyError, lambda: data.get('name'))
        self.assertEqual('foo', data.get('name', 'foo'))
        self.assertEqual({'age': 22}, data.get('person'))

    def test_things_yiled_from_data_handler(self):
        server = self.server 

        class TestSpider(Spider):
            def prepare(self):
                self.data_processed = []

            def task_page(self, grab, task):
                yield Data('foo', count=task.get('count', 1))

            def data_foo(self, count):
                self.data_processed.append(count)
                if count == 1:
                    yield Data('foo', count=666)
                    yield Task('page', url=server.get_url(),
                               count=count + 1)

        bot = build_spider(TestSpider)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertEqual(bot.data_processed, [1, 666, 2])



from grab import Grab
from grab.error import GrabTooManyRedirectsError
from test.util import BaseGrabTestCase
from test.util import GRAB_TRANSPORT


class RedirectController(object):
    def __init__(self, counter):
        self.setup_counter(counter)

    def setup_counter(self, counter):
        self.counter = counter

    def request_handler(self, server):
        if self.counter:
            server.set_status(301)
            server.set_header('Location', self.server.get_url())
        else:
            server.set_status(200)
        self.counter -= 1


class RefreshRedirectController(RedirectController):
    def request_handler(self, server):
        server.set_status(200)
        if self.counter:
            server.write('<html><head><meta '
                         'http-equiv="refresh" content="5"></head>')
        else:
            server.write('OK')
        self.counter -= 1


class GrabRedirectTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_meta_refresh_redirect(self):
        # By default meta-redirect is off
        meta_url = self.server.get_url() + '/foo'

        self.server.response_once['get.data'] =\
            '<meta http-equiv="refresh" content="5; url=%s">' % meta_url
        g = Grab(transport=GRAB_TRANSPORT)
        g.go(self.server.get_url() + '/')
        self.assertEqual(self.server.request['path'], '/')
        self.assertEqual(g.response.url, self.server.get_url() + '/')

        # Now test meta-auto-redirect
        self.server.response_once['get.data'] =\
            '<meta http-equiv="refresh" content="5; url=%s">' % meta_url
        g = Grab(transport=GRAB_TRANSPORT)
        g.setup(follow_refresh=True)
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['path'], '/foo')
        self.assertEqual(g.response.url, meta_url)

        # Test spaces in meta tag
        self.server.response_once['get.data'] =\
            "<meta http-equiv='refresh' content='0;url= %s'>" % meta_url
        g = Grab(transport=GRAB_TRANSPORT)
        g.setup(follow_refresh=True)
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['path'], '/foo')
        self.assertEqual(g.response.url, meta_url)

    def test_redirect_limit(self):
        ctl = RedirectController(10)
        self.server.response['get_callback'] = ctl.request_handler

        g = Grab(transport=GRAB_TRANSPORT)
        g.setup(redirect_limit=5)

        self.assertRaises(GrabTooManyRedirectsError,
                          lambda: g.go(self.server.get_url()))

        ctl.setup_counter(10)
        g.setup(redirect_limit=20)
        g.go(self.server.get_url())

    def test_refresh_redirect_limit(self):
        ctl = RefreshRedirectController(10)
        self.server.response['get_callback'] = ctl.request_handler

        g = Grab(transport=GRAB_TRANSPORT)
        g.setup(redirect_limit=5, follow_refresh=False)
        g.go(self.server.get_url())

        ctl.setup_counter(10)
        g.setup(redirect_limit=5, follow_refresh=True)
        self.assertRaises(GrabTooManyRedirectsError,
                          lambda: g.go(self.server.get_url()))



# coding: utf-8
import re
from weblib.error import DataNotFound
import six

from test.util import build_grab
from test.util import BaseGrabTestCase

HTML = u"""
<head>
    <title>фыва</title>
    <meta http-equiv="Content-Type" content="text/html; charset=cp1251" />
</head>
<body>
    <div id="bee">
        <div class="wrapper">
            # russian LA
            <strong id="bee-strong">пче</strong><em id="bee-em">ла</em>
        </div>
        <script type="text/javascript">
        mozilla = 777;
        </script>
        <style type="text/css">
        body { color: green; }
        </style>
    </div>
    <div id="fly">
        # russian XA
        <strong id="fly-strong">му\n</strong><em id="fly-em">ха</em>
    </div>
    <ul id="num">
        <li id="num-1">item #100 2</li>
        <li id="num-2">item #2</li>
    </ul>
""".encode('cp1251')


class ExtensionRexTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

        # Create fake grab instance with fake response
        self.g = build_grab()
        self.g.fake_response(HTML, charset='cp1251')

    def test_rex(self):
        # Search unicode rex in unicode body - default case
        rex = re.compile(u'(фыва)', re.U)
        self.assertEqual(u'фыва', self.g.rex(rex).group(1))

        # Search non-unicode rex in byte-string body
        rex = re.compile(u'(фыва)'.encode('cp1251'))
        self.assertEqual(u'фыва'.encode('cp1251'),
                         self.g.rex(rex, byte=True).group(1))

        # # Search for non-unicode rex in unicode body should fail
        pattern = '(фыва)'
        # py3 hack
        if six.PY3:
            pattern = pattern.encode('utf-8')
        rex = re.compile(pattern)
        self.assertRaises(DataNotFound, lambda: self.g.rex(rex))

        # # Search for unicode rex in byte-string body shuld fail
        rex = re.compile(u'фыва', re.U)
        self.assertRaises(DataNotFound, lambda: self.g.rex(rex, byte=True))

        # # Search for unexesting fragment
        rex = re.compile(u'(фыва2)', re.U)
        self.assertRaises(DataNotFound, lambda: self.g.rex(rex))

    def test_assert_rex(self):
        self.g.assert_rex(re.compile(u'фыва'))
        self.g.assert_rex(re.compile(u'фыва'.encode('cp1251')), byte=True)

    def test_assert_rex_text(self):
        self.assertEqual(u'ха', self.g.rex_text('<em id="fly-em">([^<]+)'))



# coding: utf-8
from unittest import TestCase

from test.util import build_grab
from test.server import SERVER


class GrabRequestTestCase(TestCase):
    def setUp(self):
        SERVER.reset()

    def test_get_method(self):
        g = build_grab()
        g.go(SERVER.BASE_URL)
        self.assertEquals('GET', SERVER.REQUEST['method'])

    def test_delete_method(self):
        g = build_grab()
        g.setup(method='delete')
        g.go(SERVER.BASE_URL)
        self.assertEquals('DELETE', SERVER.REQUEST['method'])

    def test_put_method(self):
        g = build_grab()
        g.setup(method='put', post=b'')
        g.go(SERVER.BASE_URL)
        self.assertEquals('PUT', SERVER.REQUEST['method'])



# coding: utf-8
from test.util import build_grab
from test.util import BaseGrabTestCase


class GrabRequestTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_get_method(self):
        g = build_grab()
        g.go(self.server.get_url())
        self.assertEquals('GET', self.server.request['method'])

    def test_delete_method(self):
        g = build_grab()
        g.setup(method='delete')
        g.go(self.server.get_url())
        self.assertEquals('DELETE', self.server.request['method'])

    def test_put_method(self):
        g = build_grab()
        g.setup(method='put', post=b'abc')
        g.go(self.server.get_url())
        self.assertEquals('PUT', self.server.request['method'])
        self.assertEquals('3', self.server.request['headers']['Content-Length'])



from setuptools import setup, find_packages
import os

ROOT = os.path.dirname(os.path.realpath(__file__))

setup(
    name='grab',
    version='0.6.16',
    description='Web Scraping Framework',
    long_description=open(os.path.join(ROOT, 'README.rst')).read(),
    url='http://grablib.org',
    author='Gregory Petukhov',
    author_email='lorien@lorien.name',

    packages=find_packages(),
    include_package_data=True,
    install_requires=['lxml', 'pycurl', 'selection', 'weblib>=0.1.10', 'six'],

    license="MIT",
    keywords="pycurl multicurl curl network parsing grabbing scraping"
             " lxml xpath data mining",
    classifiers=(
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
    ),
)



# Copyright: 2013, Grigoriy Petukhov
# Author: Grigoriy Petukhov (http://lorien.name)
# License: MIT
"""
The Document class is the result of network request made with Grab instance.
"""
from __future__ import absolute_import
import weakref
import re
from copy import copy
import logging
import email
import os
import json
import tempfile
import webbrowser
import codecs
from datetime import datetime
import time
from selection import XpathSelector
import six
from six.moves.urllib.parse import urlsplit, parse_qs, urljoin

import weblib.encoding
from grab.cookie import CookieManager
from weblib.files import hashed_path
from weblib.structured import TreeInterface
from weblib.text import normalize_space
from weblib.html import decode_entities
from grab.error import GrabMisuseError, DataNotFound
from weblib.rex import normalize_regexp
from grab.const import NULL
from weblib.http import smart_urlencode

logger = logging.getLogger('grab.response')
NULL_BYTE = chr(0)
RE_XML_DECLARATION = re.compile(br'^[^<]{,100}<\?xml[^>]+\?>', re.I)
RE_DECLARATION_ENCODING = re.compile(br'encoding\s*=\s*["\']([^"\']+)["\']')
RE_META_CHARSET =\
    re.compile(br'<meta[^>]+content\s*=\s*[^>]+charset=([-\w]+)', re.I)
RE_UNICODE_XML_DECLARATION =\
    re.compile(RE_XML_DECLARATION.pattern.decode('utf-8'), re.I)

# Bom processing logic was copied from
# https://github.com/scrapy/w3lib/blob/master/w3lib/encoding.py
_BOM_TABLE = [
    (codecs.BOM_UTF32_BE, 'utf-32-be'),
    (codecs.BOM_UTF32_LE, 'utf-32-le'),
    (codecs.BOM_UTF16_BE, 'utf-16-be'),
    (codecs.BOM_UTF16_LE, 'utf-16-le'),
    (codecs.BOM_UTF8, 'utf-8')
]
_FIRST_CHARS = set(char[0] for (char, name) in _BOM_TABLE)


def read_bom(data):
    """Read the byte order mark in the text, if present, and
    return the encoding represented by the BOM and the BOM.

    If no BOM can be detected, (None, None) is returned.
    """
    # common case is no BOM, so this is fast
    if data and data[0] in _FIRST_CHARS:
        for bom, encoding in _BOM_TABLE:
            if data.startswith(bom):
                return encoding, bom
    return None, None


class TextExtension(object):
    __slots__ = ()

    def text_search(self, anchor, byte=False):
        """
        Search the substring in response body.

        :param anchor: string to search
        :param byte: if False then `anchor` should be the
            unicode string, and search will be performed in
            `response.unicode_body()` else `anchor` should be the byte-string
            and search will be performed in `response.body`

        If substring is found return True else False.
        """

        if isinstance(anchor, six.text_type):
            if byte:
                raise GrabMisuseError('The anchor should be bytes string in '
                                      'byte mode')
            else:
                return anchor in self.unicode_body()

        if not isinstance(anchor, six.text_type):
            if byte:
                # if six.PY3:
                    # return anchor in self.body_as_bytes()
                return anchor in self.body
            else:
                raise GrabMisuseError('The anchor should be byte string in '
                                      'non-byte mode')

    def text_assert(self, anchor, byte=False):
        """
        If `anchor` is not found then raise `DataNotFound` exception.
        """

        if not self.text_search(anchor, byte=byte):
            raise DataNotFound(u'Substring not found: %s' % anchor)

    def text_assert_any(self, anchors, byte=False):
        """
        If no `anchors` were found then raise `DataNotFound` exception.
        """

        found = False
        for anchor in anchors:
            if self.text_search(anchor, byte=byte):
                found = True
                break
        if not found:
            raise DataNotFound(u'Substrings not found: %s'
                               % ', '.join(anchors))


class RegexpExtension(object):
    __slots__ = ()

    def rex_text(self, regexp, flags=0, byte=False, default=NULL):
        """
        Search regular expression in response body and return content of first
        matching group.

        :param byte: if False then search is performed in
        `response.unicode_body()` else the rex is searched in `response.body`.
        """

        try:
            match = self.rex_search(regexp, flags=flags, byte=byte)
        except DataNotFound:
            if default is NULL:
                raise DataNotFound('Regexp not found')
            else:
                return default
        else:
            return normalize_space(decode_entities(match.group(1)))

    def rex_search(self, regexp, flags=0, byte=False, default=NULL):
        """
        Search the regular expression in response body.

        :param byte: if False then search is performed in
        `response.unicode_body()` else the rex is searched in `response.body`.

        Note: if you use default non-byte mode than do not forget to build your
        regular expression with re.U flag.

        Return found match object or None

        """

        regexp = normalize_regexp(regexp, flags)
        match = None
        if byte:
            if not isinstance(regexp.pattern, six.text_type) or not six.PY3:
                # if six.PY3:
                    # body = self.body_as_bytes()
                # else:
                    # body = self.body
                match = regexp.search(self.body)
        else:
            if isinstance(regexp.pattern, six.text_type) or not six.PY3:
                ubody = self.unicode_body()
                match = regexp.search(ubody)
        if match:
            return match
        else:
            if default is NULL:
                raise DataNotFound('Could not find regexp: %s' % regexp)
            else:
                return default

    def rex_assert(self, rex, byte=False):
        """
        If `rex` expression is not found then raise `DataNotFound` exception.
        """

        self.rex_search(rex, byte=byte)


class PyqueryExtension(object):
    __slots__ = ()

    @property
    def pyquery(self):
        """
        Returns pyquery handler.
        """

        if not self._pyquery:
            from pyquery import PyQuery

            self._pyquery = PyQuery(self.body)
        return self._pyquery


class BodyExtension(object):
    __slots__ = ()

    def get_body_chunk(self):
        body_chunk = None
        if self.body_path:
            with open(self.body_path, 'rb') as inp:
                body_chunk = inp.read(4096)
        elif self._cached_body:
            body_chunk = self._cached_body[:4096]
        return body_chunk

    def convert_body_to_unicode(self, body, bom, charset,
                                ignore_errors, fix_special_entities):
        # How could it be unicode???
        # if isinstance(body, unicode):
            # body = body.encode('utf-8')
        if bom:
            body = body[len(self.bom):]
        if fix_special_entities:
            body = weblib.encoding.fix_special_entities(body)
        if ignore_errors:
            errors = 'ignore'
        else:
            errors = 'strict'
        return body.decode(charset, errors).strip()

    def _check_cached_body(self):
        """
        WTF???
        """
        if not self._cached_body:
            if self.body_path:
                self._cached_body = self.read_body_from_file()

    def read_body_from_file(self):
        with open(self.body_path, 'rb') as inp:
            return inp.read()

    def unicode_body(self, ignore_errors=True, fix_special_entities=True):
        """
        Return response body as unicode string.
        """

        # self._check_cached_body()
        if not self._unicode_body:
            self._unicode_body = self.convert_body_to_unicode(
                body=self.body,  # _cached_body,
                bom=self.bom,
                charset=self.charset,
                ignore_errors=ignore_errors,
                fix_special_entities=fix_special_entities,
            )
        return self._unicode_body

    def _read_body(self):
        if self.body_path:
            return self.read_body_from_file()
        else:
            return self._cached_body

    def _write_body(self, body):
        if self.body_path:
            with open(self.body_path, 'wb') as out:
                out.write(body)
            self._cached_body = None
        else:
            self._cached_body = body
        self._unicode_body = None

    body = property(_read_body, _write_body)


class DomTreeExtension(object):
    __slots__ = ()

    @property
    def tree(self):
        """
        Return DOM tree of the document built with HTML DOM builder.
        """

        if self.grab.config['content_type'] == 'xml':
            return self.build_xml_tree()
        else:
            return self.build_html_tree()

    def build_html_tree(self):
        from lxml.html import fromstring
        from lxml.etree import ParserError

        from grab.base import GLOBAL_STATE

        if self._lxml_tree is None:
            # body = self.unicode_runtime_body(
            fix_setting = self.grab.config['fix_special_entities']
            body = self.unicode_body(fix_special_entities=fix_setting).strip()
            if self.grab.config['lowercased_tree']:
                body = body.lower()
            if self.grab.config['strip_null_bytes']:
                body = body.replace(NULL_BYTE, '')
            # py3 hack
            if six.PY3:
                body = RE_UNICODE_XML_DECLARATION.sub('', body)
            else:
                body = RE_XML_DECLARATION.sub('', body)
            if not body:
                # Generate minimal empty content
                # which will not break lxml parser
                body = '<html></html>'
            start = time.time()

            # body = simplify_html(body)
            try:
                self._lxml_tree = fromstring(body)
            except Exception as ex:
                if (isinstance(ex, ParserError)
                        and 'Document is empty' in str(ex)
                        and '<html' not in body):
                    # Fix for "just a string" body
                    body = '<html>%s</html>'.format(body)
                    self._lxml_tree = fromstring(body)

                elif (isinstance(ex, TypeError)
                      and "object of type 'NoneType' has no len" in str(ex)
                      and '<html' not in body):

                    # Fix for smth like "<frameset></frameset>"
                    body = '<html>%s</html>'.format(body)
                    self._lxml_tree = fromstring(body)
                else:
                    raise

            GLOBAL_STATE['dom_build_time'] += (time.time() - start)
        return self._lxml_tree

    @property
    def xml_tree(self):
        """
        Return DOM-tree of the document built with XML DOM builder.
        """
        logger.debug(
            'This method is deprecated. Please use `tree` property '
            'and content_type="xml" option instead.')
        return self.build_xml_tree()

    def build_xml_tree(self):
        from lxml.etree import fromstring

        if self._strict_lxml_tree is None:
            # py3 hack
            # if six.PY3:
                # body = self.body_as_bytes(encode=True)
            # else:
                # body = self.body
            self._strict_lxml_tree = fromstring(self.body)
        return self._strict_lxml_tree


class FormExtension(object):
    __slots__ = ()

    def choose_form(self, number=None, id=None, name=None, xpath=None):
        """
        Set the default form.

        :param number: number of form (starting from zero)
        :param id: value of "id" attribute
        :param name: value of "name" attribute
        :param xpath: XPath query
        :raises: :class:`DataNotFound` if form not found
        :raises: :class:`GrabMisuseError`
            if method is called without parameters

        Selected form will be available via `form` attribute of `Grab`
        instance. All form methods will work with default form.

        Examples::

            # Select second form
            g.choose_form(1)

            # Select by id
            g.choose_form(id="register")

            # Select by name
            g.choose_form(name="signup")

            # Select by xpath
            g.choose_form(xpath='//form[contains(@action, "/submit")]')
        """

        if id is not None:
            try:
                self._lxml_form = self.select('//form[@id="%s"]' % id).node()
            except IndexError:
                raise DataNotFound("There is no form with id: %s" % id)
        elif name is not None:
            try:
                self._lxml_form = self.select(
                    '//form[@name="%s"]' % name).node()
            except IndexError:
                raise DataNotFound('There is no form with name: %s' % name)
        elif number is not None:
            try:
                self._lxml_form = self.tree.forms[number]
            except IndexError:
                raise DataNotFound('There is no form with number: %s' % number)
        elif xpath is not None:
            try:
                self._lxml_form = self.select(xpath).node()
            except IndexError:
                raise DataNotFound(
                    'Could not find form with xpath: %s' % xpath)
        else:
            raise GrabMisuseError('choose_form methods requires one of '
                                  '[number, id, name, xpath] arguments')

    @property
    def form(self):
        """
        This attribute points to default form.

        If form was not selected manually then select the form
        which has the biggest number of input elements.

        The form value is just an `lxml.html` form element.

        Example::

            g.go('some URL')
            # Choose form automatically
            print g.form

            # And now choose form manually
            g.choose_form(1)
            print g.form
        """

        if self._lxml_form is None:
            forms = [(idx, len(list(x.fields)))
                     for idx, x in enumerate(self.tree.forms)]
            if len(forms):
                idx = sorted(forms, key=lambda x: x[1], reverse=True)[0][0]
                self.choose_form(idx)
            else:
                raise DataNotFound('Response does not contains any form')
        return self._lxml_form

    def set_input(self, name, value):
        """
        Set the value of form element by its `name` attribute.

        :param name: name of element
        :param value: value which should be set to element

        To check/uncheck the checkbox pass boolean value.

        Example::

            g.set_input('sex', 'male')

            # Check the checkbox
            g.set_input('accept', True)
        """

        if self._lxml_form is None:
            self.choose_form_by_element('.//*[@name="%s"]' % name)
        elem = self.form.inputs[name]

        processed = False
        if getattr(elem, 'type', None) == 'checkbox':
            if isinstance(value, bool):
                elem.checked = value
                processed = True

        if not processed:
            # We need to remember original values of file fields
            # Because lxml will convert UploadContent/UploadFile object to
            # string
            if getattr(elem, 'type', '').lower() == 'file':
                self._file_fields[name] = value
                elem.value = ''
            else:
                elem.value = value

    def set_input_by_id(self, _id, value):
        """
        Set the value of form element by its `id` attribute.

        :param _id: id of element
        :param value: value which should be set to element
        """

        xpath = './/*[@id="%s"]' % _id
        if self._lxml_form is None:
            self.choose_form_by_element(xpath)
        sel = XpathSelector(self.form)
        elem = sel.select(xpath).node()
        return self.set_input(elem.get('name'), value)

    def set_input_by_number(self, number, value):
        """
        Set the value of form element by its number in the form

        :param number: number of element
        :param value: value which should be set to element
        """

        sel = XpathSelector(self.form)
        elem = sel.select('.//input[@type="text"]')[number].node()
        return self.set_input(elem.get('name'), value)

    def set_input_by_xpath(self, xpath, value):
        """
        Set the value of form element by xpath

        :param xpath: xpath path
        :param value: value which should be set to element
        """

        elem = self.select(xpath).node()

        if self._lxml_form is None:
            # Explicitly set the default form
            # which contains found element
            parent = elem
            while True:
                parent = parent.getparent()
                if parent.tag == 'form':
                    self._lxml_form = parent
                    break

        return self.set_input(elem.get('name'), value)

    # TODO:
    # Remove set_input_by_id
    # Remove set_input_by_number
    # New method: set_input_by(id=None, number=None, xpath=None)

    def submit(self, submit_name=None, make_request=True,
               url=None, extra_post=None):
        """
        Submit default form.

        :param submit_name: name of button which should be "clicked" to
            submit form
        :param make_request: if `False` then grab instance will be
            configured with form post data but request will not be
            performed
        :param url: explicitly specify form action url
        :param extra_post: (dict or list of pairs) additional form data which
            will override data automatically extracted from the form.

        Following input elements are automatically processed:

        * input[type="hidden"] - default value
        * select: value of last option
        * radio - ???
        * checkbox - ???

        Multipart forms are correctly recognized by grab library.

        Example::

            # Assume that we going to some page with some form
            g.go('some url')
            # Fill some fields
            g.set_input('username', 'bob')
            g.set_input('pwd', '123')
            # Submit the form
            g.submit()

            # or we can just fill the form
            # and do manual submission
            g.set_input('foo', 'bar')
            g.submit(make_request=False)
            g.request()

            # for multipart forms we can specify files
            from grab import UploadFile
            g.set_input('img', UploadFile('/path/to/image.png'))
            g.submit()
        """

        # TODO: add .x and .y items
        # if submit element is image

        post = self.form_fields()

        # Build list of submit buttons which have a name
        submit_controls = {}
        for elem in self.form.inputs:
            if (elem.tag == 'input' and elem.type == 'submit' and
                    elem.get('name') is not None):
                submit_controls[elem.name] = elem

        # All this code need only for one reason:
        # to not send multiple submit keys in form data
        # in real life only this key is submitted whose button
        # was pressed
        if len(submit_controls):
            # If name of submit control is not given then
            # use the name of first submit control
            if submit_name is None or submit_name not in submit_controls:
                controls = sorted(submit_controls.values(),
                                  key=lambda x: x.name)
                submit_name = controls[0].name

            # Form data should contain only one submit control
            for name in submit_controls:
                if name != submit_name:
                    if name in post:
                        del post[name]

        if url:
            action_url = urljoin(self.url, url)
        else:
            action_url = urljoin(self.url, self.form.action)

        # Values from `extra_post` should override values in form
        # `extra_post` allows multiple value of one key

        # Process saved values of file fields
        if self.form.method == 'POST':
            if 'multipart' in self.form.get('enctype', ''):
                for key, obj in self._file_fields.items():
                    post[key] = obj

        post_items = list(post.items())
        del post

        if extra_post:
            if isinstance(extra_post, dict):
                extra_post_items = extra_post.items()
            else:
                extra_post_items = extra_post

            # Drop existing post items with such key
            keys_to_drop = set([x for x, y in extra_post_items])
            for key in keys_to_drop:
                post_items = [(x, y) for x, y in post_items if x != key]

            for key, value in extra_post_items:
                post_items.append((key, value))

        if self.form.method == 'POST':
            if 'multipart' in self.form.get('enctype', ''):
                self.grab.setup(multipart_post=post_items)
            else:
                self.grab.setup(post=post_items)
            self.grab.setup(url=action_url)

        else:
            url = action_url.split('?')[0] + '?' + smart_urlencode(post_items)
            self.grab.setup(url=url)

        if make_request:
            return self.grab.request()
        else:
            return None

    def form_fields(self):
        """
        Return fields of default form.

        Fill some fields with reasonable values.
        """

        fields = dict(self.form.fields)
        for elem in self.form.inputs:
            # Ignore elements without name
            if not elem.get('name'):
                continue

            # Do not submit disabled fields
            # http://www.w3.org/TR/html4/interact/forms.html#h-17.12
            if elem.get('disabled'):
                if elem.name in fields:
                    del fields[elem.name]

            elif elem.tag == 'select':
                if fields[elem.name] is None:
                    if len(elem.value_options):
                        fields[elem.name] = elem.value_options[0]

            elif getattr(elem, 'type', None) == 'radio':
                if fields[elem.name] is None:
                    fields[elem.name] = elem.get('value')

            elif getattr(elem, 'type', None) == 'checkbox':
                if not elem.checked:
                    if elem.name is not None:
                        if elem.name in fields:
                            del fields[elem.name]

        return fields

    def choose_form_by_element(self, xpath):
        elem = self.select(xpath).node()
        while elem is not None:
            if elem.tag == 'form':
                self._lxml_form = elem
                return
            else:
                elem = elem.getparent()
        self._lxml_form = None


class Document(TextExtension, RegexpExtension, PyqueryExtension,
               BodyExtension, DomTreeExtension, FormExtension):
    """
    Document (in most cases it is a network response
        i.e. result of network request)
    """

    __slots__ = ('status', 'code', 'head', '_cached_body', '_runtime_body',
                 'body_path', 'headers', 'url', 'cookies',
                 'charset', '_unicode_body', '_unicode_runtime_body',
                 'bom', 'timestamp',
                 'name_lookup_time', 'connect_time', 'total_time',
                 'download_size', 'upload_size', 'download_speed',
                 'error_code', 'error_msg', 'grab', 'remote_ip',
                 '_lxml_tree', '_strict_lxml_tree', '_pyquery',
                 '_lxml_form', '_file_fields', 'from_cache',
                 )

    def __init__(self, grab=None):
        if grab is None:
            self.grab = None
        else:
            if isinstance(grab, weakref.ProxyType):
                self.grab = grab
            else:
                self.grab = weakref.proxy(grab)

        self.status = None
        self.code = None
        self.head = None
        self.headers = None
        self.url = None
        self.cookies = CookieManager()
        self.charset = 'utf-8'
        self.bom = None
        self.timestamp = datetime.utcnow()
        self.name_lookup_time = 0
        self.connect_time = 0
        self.total_time = 0
        self.download_size = 0
        self.upload_size = 0
        self.download_speed = 0
        self.error_code = None
        self.error_msg = None
        self.from_cache = False

        # Body
        self.body_path = None
        self._cached_body = None
        self._unicode_body = None
        self._runtime_body = None
        self._unicode_runtime_body = None

        # DOM Tree
        self._lxml_tree = None
        self._strict_lxml_tree = None

        # Pyquery
        self._pyquery = None

        # Form
        self._lxml_form = None
        self._file_fields = {}

    def __call__(self, query):
        return self.select(query)

    def select(self, *args, **kwargs):
        return XpathSelector(self.tree).select(*args, **kwargs)

    def structure(self, *args, **kwargs):
        return TreeInterface(self.tree).structured_xpath(*args, **kwargs)

    def parse(self, charset=None):
        """
        Parse headers and cookies.

        This method is called after Grab instance performes network request.
        """

        # Extract only valid lines which contain ":" character
        valid_lines = []
        for line in self.head.split('\n'):
            line = line.rstrip('\r')
            if line:
                # Each HTTP line meand the start of new response
                # self.head could contains info about multiple responses
                # For example, then 301/302 redirect was
                # processed automatically
                # Maybe it is a bug and should be fixed
                # Anyway, we handle this issue here and save headers
                # only from last response
                if line.startswith('HTTP'):
                    self.status = line
                    valid_lines = []
                else:
                    if ':' in line:
                        valid_lines.append(line)

        self.headers = email.message_from_string('\n'.join(valid_lines))

        if charset is None:
            if isinstance(self.body, six.text_type):
                self.charset = 'utf-8'
            else:
                self.detect_charset()
        else:
            self.charset = charset.lower()

        self._unicode_body = None

    def detect_charset(self):
        """
        Detect charset of the response.

        Try following methods:
        * meta[name="Http-Equiv"]
        * XML declaration
        * HTTP Content-Type header

        Ignore unknown charsets.

        Use utf-8 as fallback charset.
        """

        charset = None

        body_chunk = self.get_body_chunk()

        if body_chunk:
            # Try to extract charset from http-equiv meta tag
            try:
                charset = RE_META_CHARSET.search(body_chunk).group(1)
            except AttributeError:
                pass

            # TODO: <meta charset="utf-8" />
            bom_enc, bom = read_bom(body_chunk)
            if bom_enc:
                charset = bom_enc
                self.bom = bom

            # Try to process XML declaration
            if not charset:
                if body_chunk.startswith(b'<?xml'):
                    match = RE_XML_DECLARATION.search(body_chunk)
                    if match:
                        enc_match = RE_DECLARATION_ENCODING.search(
                            match.group(0))
                        if enc_match:
                            charset = enc_match.group(1)

        if not charset:
            if 'Content-Type' in self.headers:
                pos = self.headers['Content-Type'].find('charset=')
                if pos > -1:
                    charset = self.headers['Content-Type'][(pos + 8):]

        if charset:
            charset = charset.lower()
            if not isinstance(charset, str):
                # Convert to unicode (py2.x) or string (py3.x)
                charset = charset.decode('utf-8')
            # Check that python knows such charset
            try:
                codecs.lookup(charset)
            except LookupError:
                logger.error('Unknown charset found: %s' % charset)
                self.charset = 'utf-8'
            else:
                self.charset = charset

    def copy(self, new_grab=None):
        """
        Clone the Response object.
        """

        if new_grab is not None:
            obj = self.__class__(self.grab)  # Response()
        else:
            obj = self.__class__(new_grab)

        copy_keys = ('status', 'code', 'head', 'body', 'total_time',
                     'connect_time', 'name_lookup_time',
                     'url', 'charset', '_unicode_body')
        for key in copy_keys:
            setattr(obj, key, getattr(self, key))

        obj.headers = copy(self.headers)
        # TODO: Maybe, deepcopy?
        obj.cookies = copy(self.cookies)

        return obj

    def save(self, path, create_dirs=False):
        """
        Save response body to file.
        """

        path_dir, path_fname = os.path.split(path)
        if not os.path.exists(path_dir):
            try:
                os.makedirs(path_dir)
            except OSError:
                pass

        with open(path, 'wb') as out:
            if isinstance(self._cached_body, six.text_type):
                out.write(self._cached_body.encode('utf-8'))
            else:
                out.write(self._cached_body)

    def save_hash(self, location, basedir, ext=None):
        """
        Save response body into file with special path
        builded from hash. That allows to lower number of files
        per directory.

        :param location: URL of file or something else. It is
            used to build the SHA1 hash.
        :param basedir: base directory to save the file. Note that
            file will not be saved directly to this directory but to
            some sub-directory of `basedir`
        :param ext: extension which should be appended to file name. The
            dot is inserted automatically between filename and extension.
        :returns: path to saved file relative to `basedir`

        Example::

            >>> url = 'http://yandex.ru/logo.png'
            >>> g.go(url)
            >>> g.response.save_hash(url, 'some_dir', ext='png')
            'e8/dc/f2918108788296df1facadc975d32b361a6a.png'
            # the file was saved to $PWD/some_dir/e8/dc/...

        TODO: replace `basedir` with two options: root and save_to. And
        returns save_to + path
        """

        if isinstance(location, six.text_type):
            location = location.encode('utf-8')
        rel_path = hashed_path(location, ext=ext)
        path = os.path.join(basedir, rel_path)
        if not os.path.exists(path):
            path_dir, path_fname = os.path.split(path)
            try:
                os.makedirs(path_dir)
            except OSError:
                pass
            with open(path, 'wb') as out:
                if isinstance(self._cached_body, six.text_type):
                    out.write(self._cached_body.encode('utf-8'))
                else:
                    out.write(self._cached_body)
        return rel_path

    @property
    def json(self):
        """
        Return response body deserialized into JSON object.
        """

        if six.PY3:
            return json.loads(self.body.decode(self.charset))
        else:
            return json.loads(self.body)

    def url_details(self):
        """
        Return result of urlsplit function applied to response url.
        """

        return urlsplit(self.url)

    def query_param(self, key):
        """
        Return value of parameter in query string.
        """

        return parse_qs(self.url_details().query)[key][0]

    def browse(self):
        """
        Save response in temporary file and open it in GUI browser.
        """

        fh, path = tempfile.mkstemp()
        self.save(path)
        webbrowser.open('file://' + path)

    @property
    def time(self):
        logger.error('Attribute Response.time is deprecated. '
                     'Use Response.total_time instead.')
        return self.total_time

    def __getstate__(self):
        """
        Reset cached lxml objects which could not be pickled.
        """
        state = {}
        for cls in type(self).mro():
            cls_slots = getattr(cls, '__slots__', ())
            for slot in cls_slots:
                if slot != '__weakref__':
                    if hasattr(self, slot):
                        state[slot] = getattr(self, slot)

        state['_lxml_tree'] = None
        state['_strict_lxml_tree'] = None
        state['_lxml_form'] = None

        # state['doc'].grab = weakref.proxy(self)

        return state

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)



# coding: utf-8
import pycurl
import json
from grab import Grab
from grab.error import GrabMisuseError
from grab.cookie import CookieManager, create_cookie
import pickle

from test.util import TMP_FILE, build_grab
from test.util import BaseGrabTestCase


class TestCookies(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_parsing_response_cookies(self):
        g = build_grab()
        self.server.response['cookies'] = {'foo': 'bar', '1': '2'}.items()
        g.go(self.server.get_url())
        self.assertEqual(g.response.cookies['foo'], 'bar')

    def test_multiple_cookies(self):
        g = build_grab()
        self.server.response['cookies'] = []
        g.setup(cookies={'foo': '1', 'bar': '2'})
        g.go(self.server.get_url())
        self.assertEqual(
            set(map(lambda item: item.strip(),
                    self.server.request['headers']['Cookie'].split('; '))),
            set(['foo=1', 'bar=2']))

    def test_session(self):
        # Test that if Grab gets some cookies from the server
        # then it sends it back
        g = build_grab()
        g.setup(reuse_cookies=True)
        self.server.response['cookies'] = {'foo': 'bar'}.items()
        g.go(self.server.get_url())
        self.assertEqual(g.response.cookies['foo'], 'bar')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['Cookie'], 'foo=bar')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['Cookie'], 'foo=bar')

        # Test reuse_cookies=False
        g = build_grab()
        g.setup(reuse_cookies=False)
        self.server.response['cookies'] = {'foo': 'baz'}.items()
        g.go(self.server.get_url())
        self.assertEqual(g.response.cookies['foo'], 'baz')
        g.go(self.server.get_url())
        self.assertTrue(len(self.server.request['cookies']) == 0)

        # Test something
        g = build_grab()
        g.setup(reuse_cookies=True)
        self.server.response['cookies'] = {'foo': 'bar'}.items()
        g.go(self.server.get_url())
        self.assertEqual(g.response.cookies['foo'], 'bar')
        g.clear_cookies()
        g.go(self.server.get_url())
        self.assertTrue(len(self.server.request['cookies']) == 0)

    def test_redirect_session(self):
        g = build_grab()
        self.server.response['cookies'] = {'foo': 'bar'}.items()
        g.go(self.server.get_url())
        self.assertEqual(g.response.cookies['foo'], 'bar')

        # Setup one-time redirect
        g = build_grab()
        self.server.response['cookies'] = {}
        self.server.response_once['headers'] = [
            ('Location', self.server.get_url()),
            ('Set-Cookie', 'foo=bar'),
        ]
        self.server.response_once['code'] = 302
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['cookies']['foo'].value, 'bar')

    def test_load_dump(self):
        g = build_grab()
        cookies = {'foo': 'bar', 'spam': 'ham'}
        g.setup(cookies=cookies)
        g.go(self.server.get_url())
        g.dump_cookies(TMP_FILE)
        self.assertEqual(set(cookies.items()),
                         set((x['name'], x['value'])
                             for x in json.load(open(TMP_FILE))))

        g = build_grab()
        cookies = {'foo': 'bar', 'spam': u'begemot'}
        g.setup(cookies=cookies)
        g.go(self.server.get_url())
        g.dump_cookies(TMP_FILE)
        self.assertEqual(set(cookies.items()),
                         set((x['name'], x['value'])
                             for x in json.load(open(TMP_FILE))))

        # Test load cookies
        g = build_grab()
        cookies = [{'name': 'foo', 'value': 'bar',
                    'domain': self.server.address},
                   {'name': 'spam', 'value': u'begemot',
                    'domain': self.server.address}]
        json.dump(cookies, open(TMP_FILE, 'w'))
        g.load_cookies(TMP_FILE)
        self.assertEqual(set(g.cookies.items()),
                         set((x['name'], x['value']) for x in cookies))

    def test_cookiefile_empty(self):
        g = build_grab()
        # Empty file should not raise Exception
        open(TMP_FILE, 'w').write('')
        g.setup(cookiefile=TMP_FILE)
        g.go(self.server.get_url())

    def test_cookiefile(self):
        g = build_grab()

        cookies = [{'name': 'spam', 'value': 'ham',
                    'domain': self.server.address}]
        json.dump(cookies, open(TMP_FILE, 'w'))

        # One cookie are sent in server reponse
        # Another cookies is passed via the `cookiefile` option
        self.server.response['cookies'] = {'godzilla': 'monkey'}.items()
        g.setup(cookiefile=TMP_FILE, debug=True)
        g.go(self.server.get_url())
        print(g.request_head)
        self.assertEqual(self.server.request['cookies']['spam'].value, 'ham')

        # This is correct reslt of combining two cookies
        MERGED_COOKIES = [('godzilla', 'monkey'), ('spam', 'ham')]

        # g.cookies should contains merged cookies
        self.assertEqual(set(MERGED_COOKIES),
                         set(g.cookies.items()))

        # `cookiefile` file should contains merged cookies
        self.assertEqual(set(MERGED_COOKIES),
                         set((x['name'], x['value'])
                             for x in json.load(open(TMP_FILE))))

        # Just ensure it works
        g.go(self.server.get_url())

    def test_manual_dns(self):
        g = build_grab()
        g.transport.curl.setopt(pycurl.RESOLVE,
                                ['foo:%d:127.0.0.1' % self.server.port])
        self.server.response['get.data'] = 'zzz'
        g.go('http://foo:%d/' % self.server.port)
        self.assertEqual(b'zzz', g.response.body)

    def test_different_domains(self):
        g = build_grab()
        names = [
            'foo:%d:127.0.0.1' % self.server.port,
            'bar:%d:127.0.0.1' % self.server.port,
        ]
        g.transport.curl.setopt(pycurl.RESOLVE, names)

        self.server.response['cookies'] = {'foo': 'foo'}.items()
        g.go('http://foo:%d' % self.server.port)
        self.assertEqual(dict(g.response.cookies.items()), {'foo': 'foo'})

        self.server.response['cookies'] = {'bar': 'bar'}.items()
        g.go('http://bar:%d' % self.server.port)

        # response.cookies contains cookies from both domains
        # because it just accumulates cookies over time
        self.assertEqual(dict(g.response.cookies.items()), {'foo': 'foo',
                                                            'bar': 'bar'})

    def test_cookie_domain(self):
        g = Grab()
        names = [
            'example.com:%d:127.0.0.1' % self.server.port,
        ]
        g.transport.curl.setopt(pycurl.RESOLVE, names)
        g.cookies.set('foo', 'bar', domain='example.com')
        g.go('http://example.com:%d/' % self.server.port)

    def test_update_invalid_cookie(self):
        g = build_grab()
        self.assertRaises(GrabMisuseError, g.cookies.update, None)
        self.assertRaises(GrabMisuseError, g.cookies.update, 'asdf')
        self.assertRaises(GrabMisuseError, g.cookies.update, ['asdf'])

    def test_from_cookie_list(self):
        cookie = create_cookie('foo', 'bar', self.server.address)
        mgr = CookieManager.from_cookie_list([cookie])
        test_cookie = [x for x in mgr.cookiejar if x.name == 'foo'][0]
        self.assertEqual(cookie.name, test_cookie.name)

        mgr = CookieManager.from_cookie_list([])
        self.assertEqual(0, len(list(mgr.cookiejar)))

    def test_pickle_serialization(self):
        cookie = create_cookie('foo', 'bar', self.server.address)
        mgr = CookieManager.from_cookie_list([cookie])
        dump = pickle.dumps(mgr)
        mgr2 = pickle.loads(dump)
        self.assertEqual(list(mgr.cookiejar)[0].value,
                         list(mgr2.cookiejar)[0].value)

    def test_get_item(self):
        cookie = create_cookie('foo', 'bar', self.server.address)
        mgr = CookieManager.from_cookie_list([cookie])
        self.assertEqual('bar', mgr['foo'])
        self.assertRaises(KeyError, lambda: mgr['zzz'])

    def test_dot_domain(self):
        g = build_grab(debug=True)
        names = [
            'foo.bar:%d:127.0.0.1' % self.server.port,
            'www.foo.bar:%d:127.0.0.1' % self.server.port,
        ]
        g.transport.curl.setopt(pycurl.RESOLVE, names)

        self.server.response['headers'] = [
            ('Set-Cookie', 'foo=foo; Domain=.foo.bar; '
                           'Expires=Wed, 13 Jan 2021 22:23:01 GMT;')
        ]

        g.go('http://www.foo.bar:%d' % self.server.port)
        self.assertEqual(dict(g.response.cookies.items()), {'foo': 'foo'})

        g.go('http://www.foo.bar:%d' % self.server.port)
        self.assertEqual('foo',
                         self.server.request['cookies'].get('foo').value)

    def test_path(self):
        self.server.response['headers'] = [
            ('Set-Cookie', 'foo=1; path=/;'),
            ('Set-Cookie', 'bar=1; path=/admin;'),
        ]

        # work with "/" path
        g = build_grab()
        # get cookies
        g.go(self.server.get_url('/'))
        # submit received cookies
        g.go(self.server.get_url('/'))
        self.assertEqual(1, len(self.server.request['cookies']))

        # work with "/admin" path
        g = build_grab()
        # get cookies
        g.go(self.server.get_url('/'))
        # submit received cookies
        g.go(self.server.get_url('/admin/zz'))
        self.assertEqual(2, len(self.server.request['cookies']))

    def test_common_case_www_domain(self):
        g = build_grab()
        names = [
            'www.foo.bar:%d:127.0.0.1' % self.server.port,
        ]
        g.transport.curl.setopt(pycurl.RESOLVE, names)

        # Cookies are set for root domain (not for www subdomain)
        self.server.response['headers'] = [
            ('Set-Cookie', 'foo=1; Domain=foo.bar;'),
            ('Set-Cookie', 'bar=2; Domain=.foo.bar;'),
        ]

        # get cookies
        g.go('http://www.foo.bar:%d' % self.server.port)
        # submit cookies
        g.go('http://www.foo.bar:%d' % self.server.port)
        self.assertEqual('1', self.server.request['cookies'].get('foo').value)
        self.assertEqual('2', self.server.request['cookies'].get('bar').value)

    def test_cookie_merging_replace_with_cookies_option(self):

        init_cookies = [{'name': 'foo', 'value': 'bar',
                         'domain': self.server.address}]
        json.dump(init_cookies, open(TMP_FILE, 'w'))

        g = build_grab(debug=True)
        g.cookies.load_from_file(TMP_FILE)

        cookies = {
            'foo': 'bar2',
            'sex': 'male',
        }

        g.setup(cookies=cookies)
        g.go(self.server.get_url())
        self.assertEqual(2, len(self.server.request['cookies'].items()))

    def test_cookie_merging_replace(self):
        g = Grab()
        g.cookies.set('foo', 'bar', self.server.address)
        g.cookies.set('foo', 'bar2', self.server.address)
        self.assertEqual(1, len(g.cookies.items()))

        g.cookies.set('foo', 'bar3', '')
        self.assertEqual(1, len(g.cookies.items()))

        g.cookies.set('foo', 'bar2', domain='ya.ru')
        self.assertEqual(2, len(g.cookies.items()))



# coding: utf-8
from grab import Grab
from test.util import BaseGrabTestCase

HTML = u"""
<head>
    <title>фыва</title>
    <meta http-equiv="Content-Type" content="text/html; charset=cp1251" />
</head>
<body>
    <div id="bee">
        <div class="wrapper">
            <strong id="bee-strong">пче</strong><em id="bee-em">ла</em>
        </div>
        <script type="text/javascript">
        mozilla = 777;
        </script>
        <style type="text/css">
        body { color: green; }
        </style>
    </div>
    <div id="fly">
        <strong id="fly-strong">му\n</strong><em id="fly-em">ха</em>
    </div>
    <ul id="num">
        <li id="num-1">item #100 2</li>
        <li id="num-2">item #2</li>
    </ul>
""".encode('cp1251')


class PyqueryExtensionTest(BaseGrabTestCase):
    def setUp(self):
        # Create fake grab instance with fake response
        self.g = Grab(HTML, charset='cp1251')

    def test_some_things(self):
        from pyquery import PyQuery

        self.assertEqual(self.g.pyquery('#num-1').text(), u'item #100 2')
        self.assertEqual(self.g.pyquery('li').filter(
            lambda x: '#2' in PyQuery(x).text()).text(), u'item #2')



# coding: utf-8
from test.util import build_grab
from test.util import BaseGrabTestCase


class TestContentLimit(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_nobody(self):
        g = build_grab()
        g.setup(nobody=True)
        self.server.response['get.data'] = 'foo'
        g.go(self.server.get_url())
        self.assertEqual(b'', g.response.body)
        self.assertTrue(len(g.response.head) > 0)

    def test_body_maxsize(self):
        g = build_grab()
        g.setup(body_maxsize=100)
        self.server.response['get.data'] = 'x' * 1024 * 1024
        g.go(self.server.get_url())
        # Should be less 50kb
        self.assertTrue(len(g.response.body) < 50000)



from __future__ import absolute_import
from grab.error import (GrabError, DataNotFound, GrabNetworkError,  # noqa
                        GrabMisuseError, GrabTimeoutError)
from grab.upload import UploadContent, UploadFile  # noqa
from grab.base import Grab  # noqa
from weblib.logs import default_logging  # noqa

__version__ = '0.6.16'



from unittest import TestCase
from grab.util.config import (update_dict, build_root_config,
                              build_spider_config)
from grab.util.default_config import DEFAULT_SPIDER_GLOBAL_CONFIG
from grab.spider import Spider

class UtilConfigTestCase(TestCase):
    def test_update_dict(self):
        self.assertEqual(
            update_dict({}, {'foo': 'new'}),
            {'foo': 'new'})
        self.assertEqual(
            update_dict({'foo': 'new'}, {}),
            {'foo': 'new'})
        self.assertEqual(
            update_dict({'foo': 'old', '1': '2'}, {'foo': 'new'}),
            {'foo': 'old', '1': '2'})
        self.assertEqual(
            update_dict({'foo': 'old', '1': '2'}, {'foo': 'new'},
                        overwrite=True),
            {'foo': 'new', '1': '2'})

    def test_build_root_config_empty_settings(self):
        self.assertRaises(AttributeError, build_root_config,
                          'test.files.settings_empty')

    def test_build_root_config_file_does_not_exist(self):
        self.assertRaises(ImportError, build_root_config,
                          'test.files.settings_zzzzzz')

    def test_build_root_config_minimal_settings(self):
        cfg = build_root_config('test.files.settings_minimal')
        self.assertEqual(cfg['global'], DEFAULT_SPIDER_GLOBAL_CONFIG)

    def test_build_root_config_overwrite(self):
        cfg = build_root_config('test.files.settings_overwrite')
        for key, val in  DEFAULT_SPIDER_GLOBAL_CONFIG.items():
            if key == 'spider_modules':
                self.assertEqual(cfg['global'][key], ['zzz'])
            else:
                self.assertEqual(cfg['global'][key], val)

    def test_build_spider_config_empty(self):
        class TestSpider(Spider): pass
        root_cfg = build_root_config('test.files.settings_minimal')
        cfg = build_spider_config(TestSpider, root_cfg)
        self.assertEqual(cfg, DEFAULT_SPIDER_GLOBAL_CONFIG)

    def test_build_spider_config_overwrite(self):
        class TestSpider(Spider): pass
        root_cfg = build_root_config('test.files.settings_test_spider')
        cfg = build_spider_config(TestSpider, root_cfg)
        for key, val in  DEFAULT_SPIDER_GLOBAL_CONFIG.items():
            if key == 'spider_modules':
                self.assertEqual(cfg[key], ['zzz'])
            elif key == 'thread_number':
                self.assertEqual(cfg[key], 777)
            else:
                self.assertEqual(cfg[key], val)

    def test_setup_spider_config(self):
        class TestSpider(Spider):
            @classmethod
            def setup_spider_config(cls, config):
                config['foo'] = 'bar'

        root_cfg = build_root_config('test.files.settings_minimal')
        cfg = build_spider_config(TestSpider, root_cfg)
        for key, val in  DEFAULT_SPIDER_GLOBAL_CONFIG.items():
            if key == 'foo':
                self.assertEqual(cfg[key], 'bar')
            else:
                self.assertEqual(cfg[key], val)



# coding: utf-8
from grab import DataNotFound, GrabMisuseError
import six

from test.util import build_grab
from test.util import BaseGrabTestCase

HTML = u"""
<head>
    <title>фыва</title>
    <meta http-equiv="Content-Type" content="text/html; charset=cp1251" />
</head>
<body>
    <div id="bee">
        <div class="wrapper">
            # russian LA
            <strong id="bee-strong">пче</strong><em id="bee-em">ла</em>
        </div>
        <script type="text/javascript">
        mozilla = 777;
        </script>
        <style type="text/css">
        body { color: green; }
        </style>
    </div>
    <div id="fly">
        # russian XA
        <strong id="fly-strong">му\n</strong><em id="fly-em">ха</em>
    </div>
    <ul id="num">
        <li id="num-1">item #100 2</li>
        <li id="num-2">item #2</li>
    </ul>
""".encode('cp1251')


class TextExtensionTest(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

        # Create fake grab instance with fake response
        self.g = build_grab()
        self.g.fake_response(HTML, charset='cp1251')

    def test_search(self):
        self.assertTrue(self.g.search(u'фыва'.encode('cp1251'), byte=True))
        self.assertTrue(self.g.search(u'фыва'))
        self.assertFalse(self.g.search(u'фыва2'))

    def test_search_usage_errors(self):
        self.assertRaises(GrabMisuseError, self.g.search, u'фыва', byte=True)
        anchor = 'фыва'
        # py3 hack
        if six.PY3:
            anchor = anchor.encode('utf-8')
        self.assertRaises(GrabMisuseError, self.g.search, anchor)

    def test_assert_substring(self):
        self.g.assert_substring(u'фыва')
        self.g.assert_substring(u'фыва'.encode('cp1251'), byte=True)
        self.assertRaises(DataNotFound, self.g.assert_substring, u'фыва2')

    def test_assert_substrings(self):
        self.g.assert_substrings((u'фыва',))
        self.g.assert_substrings((u'фывы нет', u'фыва'))
        self.g.assert_substrings((u'фыва'.encode('cp1251'), 'где ты фыва?'),
                                 byte=True)
        self.assertRaises(DataNotFound, self.g.assert_substrings,
                          (u'фыва, вернись', u'фыва-а-а-а'))
